@@ -27,6 +27,7 @@
 #include <QIcon>
 #include <QQmlComponent>
 #include <QCoreApplication>
+#include <QKeySequence>
 
 extern "C" {
   #include "gtk/gtkactionmuxer.h"
@@ -53,7 +54,9 @@ enum MenuRoles {
     ActionStateRole,
     IsCheckRole,
     IsRadioRole,
-    IsToggledRole
+    IsToggledRole,
+    ShortcutRole,
+    HasSubmenuRole
 };
 
 class UnityMenuModelPrivate
@@ -86,8 +89,8 @@ public:
 
     static void nameAppeared(GDBusConnection *connection, const gchar *name, const gchar *owner, gpointer user_data);
     static void nameVanished(GDBusConnection *connection, const gchar *name, gpointer user_data);
-    static void menuItemInserted(GtkMenuTrackerItem *item, gint position, gpointer user_data);
-    static void menuItemRemoved(gint position, gpointer user_data);
+    static void menuItemInserted(GPtrArray *items, gint position, gpointer user_data);
+    static void menuItemRemoved(gint position, gint n_items, gpointer user_data);
     static void menuItemChanged(GObject *object, GParamSpec *pspec, gpointer user_data);
 
     static void registeredActionAdded(GtkSimpleActionObserver    *observer_item,
@@ -247,19 +250,19 @@ void UnityMenuModelPrivate::nameVanished(GDBusConnection *connection, const gcha
     priv->clearName();
 }
 
-void UnityMenuModelPrivate::menuItemInserted(GtkMenuTrackerItem *item, gint position, gpointer user_data)
+void UnityMenuModelPrivate::menuItemInserted(GPtrArray *items, gint position, gpointer user_data)
 {
     UnityMenuModelPrivate *priv = (UnityMenuModelPrivate *)user_data;
 
-    UnityMenuModelAddRowEvent ummare(item, position);
+    UnityMenuModelAddRowEvent ummare(items, position);
     QCoreApplication::sendEvent(priv->model, &ummare);
 }
 
-void UnityMenuModelPrivate::menuItemRemoved(gint position, gpointer user_data)
+void UnityMenuModelPrivate::menuItemRemoved(gint position, gint n_items, gpointer user_data)
 {
     UnityMenuModelPrivate *priv = (UnityMenuModelPrivate *)user_data;
 
-    UnityMenuModelRemoveRowEvent ummrre(position);
+    UnityMenuModelRemoveRowEvent ummrre(position, n_items);
     QCoreApplication::sendEvent(priv->model, &ummrre);
 }
 
@@ -286,7 +289,7 @@ UnityMenuModel::UnityMenuModel(QObject *parent):
     priv = new UnityMenuModelPrivate(this);
 }
 
-UnityMenuModel::UnityMenuModel(const UnityMenuModelPrivate& other, QObject *parent):
+UnityMenuModel::UnityMenuModel(const UnityMenuModelPrivate& other, UnityMenuModel *parent):
     QAbstractListModel(parent)
 {
     priv = new UnityMenuModelPrivate(other, this);
@@ -483,7 +486,13 @@ QVariant UnityMenuModel::data(const QModelIndex &index, int role) const
             return gtk_menu_tracker_item_get_role (item) == GTK_MENU_TRACKER_ITEM_ROLE_RADIO;
 
         case IsToggledRole:
-            return gtk_menu_tracker_item_get_toggled (item) == TRUE ? true : false;
+            return gtk_menu_tracker_item_get_toggled (item) != FALSE;
+
+        case ShortcutRole:
+            return QKeySequence(gtk_menu_tracker_item_get_accel (item), QKeySequence::NativeText);
+
+        case HasSubmenuRole:
+            return gtk_menu_tracker_item_get_has_submenu (item) != FALSE;
 
         default:
             return QVariant();
@@ -500,7 +509,6 @@ QModelIndex UnityMenuModel::parent(const QModelIndex &index) const
     return QModelIndex();
 }
 
-#include <QtDebug>
 QHash<int, QByteArray> UnityMenuModel::roleNames() const
 {
     QHash<int, QByteArray> names;
@@ -516,6 +524,8 @@ QHash<int, QByteArray> UnityMenuModel::roleNames() const
     names[IsCheckRole] = "isCheck";
     names[IsRadioRole] = "isRadio";
     names[IsToggledRole] = "isToggled";
+    names[ShortcutRole] = "shortcut";
+    names[HasSubmenuRole] = "hasSubmenu";
 
     return names;
 }
@@ -661,7 +671,7 @@ bool UnityMenuModel::loadExtendedAttributes(int position, const QVariantMap &sch
             continue;
         }
 
-        QVariant qvalue = attributeToQVariant(value, type);
+        const QVariant &qvalue = attributeToQVariant(value, type);
         if (qvalue.isValid())
             extendedAttrs->insert(qtify_name (name.toUtf8()), qvalue);
         else
@@ -675,6 +685,7 @@ bool UnityMenuModel::loadExtendedAttributes(int position, const QVariantMap &sch
                              extendedAttrs, freeExtendedAttrs);
 
     Q_EMIT dataChanged(index(position, 0), index(position, 0), QVector<int>() << ExtendedAttributesRole);
+    return true;
 }
 
 QVariant UnityMenuModel::get(int row, const QByteArray &role)
@@ -692,6 +703,8 @@ void UnityMenuModel::activate(int index, const QVariant& parameter)
 {
     GSequenceIter *it;
     GtkMenuTrackerItem *item;
+    GVariant *value;
+    const GVariantType *parameter_type;
 
     it = g_sequence_get_iter_at_pos (priv->items, index);
     if (g_sequence_iter_is_end (it)) {
@@ -707,7 +720,9 @@ void UnityMenuModel::activate(int index, const QVariant& parameter)
         gchar *action;
 
         action = gtk_menu_tracker_item_get_action_name (item);
-        g_action_group_activate_action (G_ACTION_GROUP (priv->muxer), action, Converter::toGVariant(parameter));
+        parameter_type = g_action_group_get_action_parameter_type (G_ACTION_GROUP (priv->muxer), action);
+        value = Converter::toGVariantWithSchema(parameter, g_variant_type_peek_string (parameter_type));
+        g_action_group_activate_action (G_ACTION_GROUP (priv->muxer), action, value);
 
         g_free (action);
     } else {
@@ -715,6 +730,57 @@ void UnityMenuModel::activate(int index, const QVariant& parameter)
     }
 }
 
+void UnityMenuModel::aboutToShow(int index)
+{
+    GSequenceIter *it = g_sequence_get_iter_at_pos (priv->items, index);
+    if (g_sequence_iter_is_end (it)) {
+        return;
+    }
+
+    auto item = static_cast<GtkMenuTrackerItem*>(g_sequence_get(it));
+    if (!item) {
+        return;
+    }
+
+    quint64 actionTag;
+    if (gtk_menu_tracker_item_get_attribute (item, "qtubuntu-tag", "t", &actionTag)) {
+        // Child UnityMenuModel have priv->connection null, so climb to the parent until we find a non null one
+        UnityMenuModelPrivate *privToUse = priv;
+        while (privToUse && !privToUse->connection) {
+            auto pModel = dynamic_cast<UnityMenuModel*>(privToUse->model->QObject::parent());
+            if (pModel) {
+                privToUse = pModel->priv;
+            } else {
+                privToUse = nullptr;
+            }
+        }
+        if (privToUse) {
+            g_dbus_connection_call (privToUse->connection,
+                                    privToUse->busName,
+                                    privToUse->menuObjectPath,
+                                    "qtubuntu.actions.extra",
+                                    "aboutToShow",
+                                    g_variant_new("(t)", actionTag),
+                                    nullptr,
+                                    G_DBUS_CALL_FLAGS_NO_AUTO_START,
+                                    G_MAXINT,
+                                    nullptr,
+                                    nullptr,
+                                    nullptr);
+        }
+    }
+}
+
+void UnityMenuModel::activateByVariantString(int index, const QString& parameter)
+{
+    activate(index, Converter::toQVariantFromVariantString(parameter));
+}
+
+void UnityMenuModel::changeStateByVariantString(int index, const QString& parameter)
+{
+    changeState(index, Converter::toQVariantFromVariantString(parameter));
+}
+
 void UnityMenuModel::changeState(int index, const QVariant& parameter)
 {
     GSequenceIter *it;
@@ -773,26 +839,29 @@ bool UnityMenuModel::event(QEvent* e)
         GSequenceIter *it;
         it = g_sequence_get_iter_at_pos (priv->items, ummrce->position);
 
-        beginInsertRows(QModelIndex(), ummrce->position, ummrce->position);
+        beginInsertRows(QModelIndex(), ummrce->position, ummrce->position + ummrce->items->len - 1);
 
-        it = g_sequence_insert_before (it, g_object_ref (ummrce->item));
-        g_object_set_qdata (G_OBJECT (ummrce->item), unity_menu_model_quark (), this);
-        g_signal_connect (ummrce->item, "notify", G_CALLBACK (UnityMenuModelPrivate::menuItemChanged), it);
+        for (gint i = ummrce->items->len - 1; i >= 0; --i) {
+            GtkMenuTrackerItem *item = (GtkMenuTrackerItem*)g_ptr_array_index(ummrce->items, i);
+            it = g_sequence_insert_before (it, g_object_ref (item));
+            g_object_set_qdata (G_OBJECT (item), unity_menu_model_quark (), this);
+            g_signal_connect (item, "notify", G_CALLBACK (UnityMenuModelPrivate::menuItemChanged), it);
+        }
 
         endInsertRows();
         return true;
     } else if (e->type() == UnityMenuModelRemoveRowEvent::eventType) {
         UnityMenuModelRemoveRowEvent *ummrre = static_cast<UnityMenuModelRemoveRowEvent*>(e);
 
-        GSequenceIter *it;
-        it = g_sequence_get_iter_at_pos (priv->items, ummrre->position);
-        if (!g_sequence_iter_is_end (it)) {
-            beginRemoveRows(QModelIndex(), ummrre->position, ummrre->position);
-
-            g_sequence_remove (it);
-
-            endRemoveRows();
+        beginRemoveRows(QModelIndex(), ummrre->position, ummrre->position + ummrre->nItems - 1);
+        for (int i = 0; i < ummrre->nItems; ++i) {
+            GSequenceIter *it = g_sequence_get_iter_at_pos (priv->items, ummrre->position);
+            if (!g_sequence_iter_is_end (it)) {
+                g_sequence_remove (it);
+            }
         }
+        endRemoveRows();
+
         return true;
     } else if (e->type() == UnityMenuModelDataChangeEvent::eventType) {
         UnityMenuModelDataChangeEvent *ummdce = static_cast<UnityMenuModelDataChangeEvent*>(e);
